@@ -1,42 +1,48 @@
+from functools import partial
 from typing import (Iterable,
                     List,
                     Optional,
                     Sequence,
                     Tuple)
 
-from orient.planar import (PointLocation,
-                           contours_in_contour,
+from orient.planar import (contour_in_polygon,
                            point_in_polygon,
-                           polygon_in_polygon)
+                           polygon_in_polygon,
+                           region_in_multiregion,
+                           segment_in_polygon)
 from reprit.base import generate_repr
+from sect.decomposition import polygon_trapezoidal
 from sect.triangulation import constrained_delaunay_triangles
 
 from .angular import (Orientation,
                       to_orientation)
+from .compound import (Compound,
+                       Indexable,
+                       Relation,
+                       Shaped)
 from .geometry import Geometry
 from .hints import Coordinate
 from .linear import (Contour,
                      RawContour,
-                     forms_convex_polygon,
-                     to_signed_area)
-from .primitive import Point
+                     Segment,
+                     vertices)
+from .primitive import (Point,
+                        RawPoint)
 
 RawPolygon = Tuple[RawContour, List[RawContour]]
 
 
-class Polygon(Geometry):
-    __slots__ = ('_border', '_holes', '_raw_border', '_raw_holes',
-                 '_normalized_border', '_normalized_holes', '_is_normalized')
+class Polygon(Indexable, Shaped):
+    __slots__ = ('_border', '_holes', '_holes_set',
+                 '_raw_border', '_raw_holes', '_contains')
 
     def __init__(self, border: Contour,
-                 holes: Optional[Sequence[Contour]] = None,
-                 *,
-                 _is_normalized: bool = False) -> None:
+                 holes: Optional[Sequence[Contour]] = None) -> None:
         """
         Initializes polygon.
 
         Time complexity:
-            ``O(vertices_count + len(self.holes) * log len(self.holes))``
+            ``O(vertices_count)``
         Memory complexity:
             ``O(vertices_count)``
 
@@ -44,23 +50,18 @@ class Polygon(Geometry):
  + sum(len(hole.vertices) for hole in holes)``.
         """
         holes = tuple(holes or ())
-        self._border, self._holes = border, holes
+        self._border, self._holes, self._holes_set = (border, holes,
+                                                      frozenset(holes))
         self._raw_border, self._raw_holes = border.raw(), [hole.raw()
                                                            for hole in holes]
-        self._normalized_border, self._normalized_holes = (
-            (border, holes)
-            if _is_normalized
-            else (border.normalized.to_counterclockwise(),
-                  tuple(sorted([hole.normalized.to_clockwise()
-                                for hole in holes],
-                               key=lambda contour: contour._vertices[:2]))))
-        self._is_normalized = _is_normalized
+        self._contains = partial(_plain_contains,
+                                 (self._raw_border, self._raw_holes))
 
     __repr__ = generate_repr(__init__)
 
-    def __contains__(self, point: Point) -> bool:
+    def __contains__(self, other: Geometry) -> bool:
         """
-        Checks if the point lies inside the polygon or on its boundary.
+        Checks if the polygon contains the other geometry.
 
         Time complexity:
             ``O(vertices_count)``
@@ -89,9 +90,7 @@ class Polygon(Geometry):
         >>> Point(7, 0) in polygon
         False
         """
-        return (point_in_polygon(point.raw(),
-                                 (self._raw_border, self._raw_holes))
-                is not PointLocation.EXTERNAL)
+        return isinstance(other, Point) and self._contains(other.raw())
 
     def __eq__(self, other: 'Polygon') -> bool:
         """
@@ -121,75 +120,58 @@ class Polygon(Geometry):
         True
         """
         return (self is other
-                or (self._normalized_border == other._normalized_border
-                    and self._normalized_holes == other._normalized_holes
+                or (self._border == other._border
+                    and self._holes_set == other._holes_set
                     if isinstance(other, Polygon)
-                    else NotImplemented))
+                    else (False
+                          if isinstance(other, Geometry)
+                          else NotImplemented)))
 
-    def __ge__(self, other: 'Polygon') -> bool:
+    def __ge__(self, other: Compound) -> bool:
         """
-        Checks if the polygon is a superset of the other.
+        Checks if the polygon is a superset of the other geometry.
 
         Time complexity:
-            ``O(total_vertices_count * log total_vertices_count)``
+            ``O(vertices_count * log vertices_count)``
         Memory complexity:
-            ``O(total_vertices_count)``
+            ``O(1)``
 
-        where ``vertices_count = total_border_vertices_count\
- + total_holes_vertices_count``,
-        ``total_border_vertices_count =\
- border_vertices_count + other_border_vertices_count``
-        ``total_holes_vertices_count =\
- holes_vertices_count + other_holes_vertices_count``,
-        ``border_vertices_count = len(self.border.vertices)``,
-        ``other_border_vertices_count = len(other.border.vertices)``
-        ``holes_vertices_count =\
- sum(len(hole.vertices) for hole in self.holes)``,
-        ``other_holes_vertices_count =\
- sum(len(hole.vertices) for hole in other.holes)``.
+        where ``vertices_count = len(self.border.vertices)\
+ + sum(len(hole.vertices) for hole in self.holes)``.
 
         >>> polygon = Polygon.from_raw(([(0, 0), (6, 0), (6, 6), (0, 6)],
         ...                             [[(2, 2), (2, 4), (4, 4), (4, 2)]]))
         >>> polygon >= polygon
         True
-        >>> polygon >= polygon.convex_hull
-        False
         """
-        return (polygon_in_polygon((other._raw_border, other._raw_holes),
-                                   (self._raw_border, self._raw_holes))
-                if isinstance(other, Polygon)
-                else NotImplemented)
+        return (self == other
+                or (self.relate(other) in (Relation.EQUAL, Relation.COMPONENT,
+                                           Relation.ENCLOSED, Relation.WITHIN)
+                    if isinstance(other, Compound)
+                    else NotImplemented))
 
-    def __gt__(self, other: 'Polygon') -> bool:
+    def __gt__(self, other: Compound) -> bool:
         """
-        Checks if the polygon is a strict superset of the other.
+        Checks if the polygon is a strict superset of the other geometry.
 
         Time complexity:
-            ``O(total_vertices_count * log total_vertices_count)``
+            ``O(vertices_count * log vertices_count)``
         Memory complexity:
-            ``O(total_vertices_count)``
+            ``O(1)``
 
-        where ``vertices_count = total_border_vertices_count\
- + total_holes_vertices_count``,
-        ``total_border_vertices_count =\
- border_vertices_count + other_border_vertices_count``
-        ``total_holes_vertices_count =\
- holes_vertices_count + other_holes_vertices_count``,
-        ``border_vertices_count = len(self.border.vertices)``,
-        ``other_border_vertices_count = len(other.border.vertices)``
-        ``holes_vertices_count =\
- sum(len(hole.vertices) for hole in self.holes)``,
-        ``other_holes_vertices_count =\
- sum(len(hole.vertices) for hole in other.holes)``.
+        where ``vertices_count = len(self.border.vertices)\
+ + sum(len(hole.vertices) for hole in self.holes)``.
 
         >>> polygon = Polygon.from_raw(([(0, 0), (6, 0), (6, 6), (0, 6)],
         ...                             [[(2, 2), (2, 4), (4, 4), (4, 2)]]))
         >>> polygon > polygon
         False
-        >>> polygon > polygon.convex_hull
-        False
         """
-        return self != other and self >= other
+        return (self != other
+                and (self.relate(other) in (Relation.COMPONENT,
+                                            Relation.ENCLOSED, Relation.WITHIN)
+                     if isinstance(other, Compound)
+                     else NotImplemented))
 
     def __hash__(self) -> int:
         """
@@ -208,72 +190,60 @@ class Polygon(Geometry):
         >>> hash(polygon) == hash(polygon)
         True
         """
-        return hash((self._normalized_border, self._normalized_holes))
+        return hash((self._border, self._holes_set))
 
-    def __le__(self, other: 'Polygon') -> bool:
+    def __le__(self, other: Compound) -> bool:
         """
-        Checks if the polygon is a subset of the other.
+        Checks if the polygon is a subset of the other geometry.
 
         Time complexity:
-            ``O(total_vertices_count * log total_vertices_count)``
+            ``O(vertices_count * log vertices_count)``
         Memory complexity:
-            ``O(total_vertices_count)``
+            ``O(1)``
 
-        where ``vertices_count = total_border_vertices_count\
- + total_holes_vertices_count``,
-        ``total_border_vertices_count =\
- border_vertices_count + other_border_vertices_count``
-        ``total_holes_vertices_count =\
- holes_vertices_count + other_holes_vertices_count``,
-        ``border_vertices_count = len(self.border.vertices)``,
-        ``other_border_vertices_count = len(other.border.vertices)``
-        ``holes_vertices_count =\
- sum(len(hole.vertices) for hole in self.holes)``,
-        ``other_holes_vertices_count =\
- sum(len(hole.vertices) for hole in other.holes)``.
+        where ``vertices_count = len(self.border.vertices)\
+ + sum(len(hole.vertices) for hole in self.holes)``.
 
         >>> polygon = Polygon.from_raw(([(0, 0), (6, 0), (6, 6), (0, 6)],
         ...                             [[(2, 2), (2, 4), (4, 4), (4, 2)]]))
         >>> polygon <= polygon
         True
-        >>> polygon <= polygon.convex_hull
-        True
         """
-        return (polygon_in_polygon((self._raw_border, self._raw_holes),
-                                   (other._raw_border, other._raw_holes))
-                if isinstance(other, Polygon)
-                else NotImplemented)
+        return (self == other
+                or ((self.relate(other) in (Relation.COVER, Relation.ENCLOSES,
+                                            Relation.COMPOSITE, Relation.EQUAL)
+                     if isinstance(other, Shaped)
+                     # shaped cannot be subset of linear
+                     else False)
+                    if isinstance(other, Compound)
+                    else NotImplemented))
 
-    def __lt__(self, other: 'Polygon') -> bool:
+    def __lt__(self, other: Compound) -> bool:
         """
-        Checks if the polygon is a strict subset of the other.
+        Checks if the polygon is a strict subset of the other geometry.
 
         Time complexity:
-            ``O(total_vertices_count * log total_vertices_count)``
+            ``O(vertices_count * log vertices_count)``
         Memory complexity:
-            ``O(total_vertices_count)``
+            ``O(1)``
 
-        where ``vertices_count = total_border_vertices_count\
- + total_holes_vertices_count``,
-        ``total_border_vertices_count =\
- border_vertices_count + other_border_vertices_count``
-        ``total_holes_vertices_count =\
- holes_vertices_count + other_holes_vertices_count``,
-        ``border_vertices_count = len(self.border.vertices)``,
-        ``other_border_vertices_count = len(other.border.vertices)``
-        ``holes_vertices_count =\
- sum(len(hole.vertices) for hole in self.holes)``,
-        ``other_holes_vertices_count =\
- sum(len(hole.vertices) for hole in other.holes)``.
+        where ``vertices_count = len(self.border.vertices)\
+ + sum(len(hole.vertices) for hole in self.holes)``.
 
         >>> polygon = Polygon.from_raw(([(0, 0), (6, 0), (6, 6), (0, 6)],
         ...                             [[(2, 2), (2, 4), (4, 4), (4, 2)]]))
         >>> polygon < polygon
         False
-        >>> polygon < polygon.convex_hull
-        True
         """
-        return self != other and self <= other
+        return (self != other
+                and ((self.relate(other) in (Relation.COVER,
+                                             Relation.ENCLOSES,
+                                             Relation.COMPOSITE)
+                      if isinstance(other, Shaped)
+                      # shaped cannot be strict subset of linear
+                      else False)
+                     if isinstance(other, Compound)
+                     else NotImplemented))
 
     @classmethod
     def from_raw(cls, raw: RawPolygon) -> 'Polygon':
@@ -320,8 +290,9 @@ class Polygon(Geometry):
         >>> polygon.area == 32
         True
         """
-        return (to_signed_area(self._normalized_border)
-                + sum(to_signed_area(hole) for hole in self._normalized_holes))
+        return (abs(vertices.region_signed_area(self._border._vertices))
+                - sum(abs(vertices.region_signed_area(hole._vertices))
+                      for hole in self._holes))
 
     @property
     def border(self) -> Contour:
@@ -398,30 +369,48 @@ class Polygon(Geometry):
         >>> polygon.convex_hull.is_convex
         True
         """
-        return not self._holes and forms_convex_polygon(self._border)
+        return (not self._holes
+                and vertices.form_convex_polygon(self._border._vertices))
 
     @property
-    def normalized(self) -> 'Polygon':
+    def perimeter(self) -> Coordinate:
         """
-        Returns polygon in normalized form.
+        Returns perimeter of the polygon.
 
         Time complexity:
-            ``O(1)`` if normalized already, ``O(vertices_count)`` -- otherwise
+            ``O(vertices_count)``
         Memory complexity:
-            ``O(1)`` if normalized already, ``O(vertices_count)`` -- otherwise
+            ``O(1)``
 
         where ``vertices_count = len(self.border.vertices)\
  + sum(len(hole.vertices) for hole in self.holes)``.
 
         >>> polygon = Polygon.from_raw(([(0, 0), (6, 0), (6, 6), (0, 6)],
         ...                             [[(2, 2), (2, 4), (4, 4), (4, 2)]]))
-        >>> polygon.normalized == polygon
+        >>> polygon.perimeter == 32
         True
         """
-        return (self
-                if self._is_normalized
-                else Polygon(self._normalized_border, self._normalized_holes,
-                             _is_normalized=True))
+        return self._border.length + sum(hole.length for hole in self._holes)
+
+    def index(self) -> None:
+        """
+        Pre-processes contour to potentially improve queries time complexity.
+
+        Time complexity:
+            ``O(vertices_count * log vertices_count)`` expected,
+            ``O(vertices_count ** 2)`` worst
+        Memory complexity:
+            ``O(vertices_count)``
+
+        where ``vertices_count = len(self.border.vertices)\
+ + sum(len(hole.vertices) for hole in self.holes)``.
+
+        >>> polygon = Polygon.from_raw(([(0, 0), (6, 0), (6, 6), (0, 6)],
+        ...                             [[(2, 2), (2, 4), (4, 4), (4, 2)]]))
+        >>> polygon.index()
+        """
+        graph = polygon_trapezoidal(self._raw_border, self._raw_holes)
+        self._contains = graph.__contains__
 
     def raw(self) -> RawPolygon:
         """
@@ -435,13 +424,39 @@ class Polygon(Geometry):
         where ``vertices_count = len(self.border.vertices)\
  + sum(len(hole.vertices) for hole in self.holes)``.
 
-        >>> contour = Contour.from_raw([(0, 0), (1, 0), (0, 1)])
-        >>> contour.raw()
-        [(0, 0), (1, 0), (0, 1)]
+        >>> polygon = Polygon.from_raw(([(0, 0), (6, 0), (6, 6), (0, 6)],
+        ...                             [[(2, 2), (2, 4), (4, 4), (4, 2)]]))
+        >>> polygon.raw()
+        ([(0, 0), (6, 0), (6, 6), (0, 6)], [[(2, 2), (2, 4), (4, 4), (4, 2)]])
         """
         return self._raw_border[:], [raw_hole[:]
                                      for raw_hole in self._raw_holes]
 
+    def relate(self, other: Compound) -> Relation:
+        """
+        Finds relation between the polygon and the other geometric object.
+
+        Time complexity:
+            ``O(vertices_count * log vertices_count)``
+        Memory complexity:
+            ``O(vertices_count)``
+
+        where ``vertices_count = len(self.border.vertices)\
+ + sum(len(hole.vertices) for hole in self.holes)``.
+
+        >>> polygon = Polygon.from_raw(([(0, 0), (6, 0), (6, 6), (0, 6)],
+        ...                             [[(2, 2), (2, 4), (4, 4), (4, 2)]]))
+        >>> polygon.relate(polygon) is Relation.EQUAL
+        True
+        """
+        raw = self._raw_border, self._raw_holes
+        return (segment_in_polygon(other.raw(), raw)
+                if isinstance(other, Segment)
+                else (contour_in_polygon(other.raw(), raw)
+                      if isinstance(other, Contour)
+                      else polygon_in_polygon((other._raw_border,
+                                               other._raw_holes), raw)))
+
     def triangulate(self) -> List['Polygon']:
         """
         Returns triangulation of the polygon.
@@ -488,10 +503,13 @@ class Polygon(Geometry):
         >>> polygon.validate()
         """
         self._border.validate()
-        for hole in self._holes:
-            hole.validate()
-        if not contours_in_contour(self._raw_holes, self._raw_border):
-            raise ValueError('Holes should lie inside border.')
+        if self._holes:
+            for hole in self._holes:
+                hole.validate()
+            relation = region_in_multiregion(self._raw_border, self._raw_holes)
+            if (relation is not Relation.COVER
+                    and relation is not Relation.ENCLOSES):
+                raise ValueError('Holes should lie inside border.')
 
 
 def _to_convex_hull(points: Sequence[Point]) -> List[Point]:
@@ -512,3 +530,7 @@ def _to_sub_hull(points: Iterable[Point]) -> List[Point]:
                 break
         result.append(point)
     return result
+
+
+def _plain_contains(raw_polygon: RawPolygon, raw_point: RawPoint) -> bool:
+    return bool(point_in_polygon(raw_point, raw_polygon))
